import os
import json

import click
import lkml
import yaml

from .lookml_project import LookMLProjectConverter
from .lookml_models import LookMLView, LookMLModel


def _convert_logic(path: str, directory: str, out_directory: str, convert_type: str):
    abs_path = os.path.abspath(path)
    with open(abs_path, "r") as file:
        if convert_type == "dashboard":
            lkml_result = yaml.safe_load(file)
        else:
            lkml_result = lkml.load(file)

    if directory:
        object_names = []
        to_loop = [lkml_result] if convert_type == "model" else lkml_result
        for o in lkml_result:
            if convert_type == "model":
                o["name"] = path.split(".")[-3].split("/")[-1]
            key = "name" if convert_type != "dashboard" else "dashboard"
            object_names.append(o[key])
        project = LookMLProjectConverter(in_directory=directory)
        lookml_project = project.load(project.in_directory)
        models, views, dashboards, topics = project.convert_project(lookml_project)
        if convert_type == "dashboard":
            objects = dashboards
        elif convert_type == "model":
            objects = models
        elif convert_type == "view":
            objects = views
        elif convert_type == "topic":
            objects = topics
        else:
            raise ValueError(f"Unknown convert_type {convert_type}")
        result_objects = [o for o in objects if o["name"] in object_names]
    else:
        if convert_type == "dashboard":
            raise ValueError("Dashboards need context from the --directory argument to successfully convert")
        else:
            result_objects = []
            to_loop = lkml_result["views"] if convert_type == "view" else [lkml_result]
            for o in to_loop:
                if convert_type == "model":
                    o["name"] = path.split(".")[-3].split("/")[-1]
                    model = LookMLProjectConverter().convert_model(LookMLModel.from_dict(o))
                    result_objects.append(model)
                elif convert_type == "view":
                    result_objects.append(
                        LookMLProjectConverter().convert_view(
                            LookMLView.from_dict(o), model_name="TODO add model name"
                        )
                    )

    if out_directory:
        for o in result_objects:
            if convert_type == "model":
                ext = "_model.yml"
            elif convert_type == "view":
                ext = "_view.yml"
            elif convert_type == "topic":
                ext = "_topic.yml"
            else:
                ext = ".yml"
            with open(os.path.join(os.path.abspath(out_directory), f'{o["name"]}{ext}'), "w") as f:
                if convert_type == "topic":
                    o.pop("name")
                yaml.dump(o, f)
    else:
        return result_objects


@click.group()
@click.version_option()
def cli_group():
    pass


@cli_group.command()
@click.option("--directory", default=None, help="The directory of the LookML project to use in conversion")
@click.option("--out-directory", default=None, help="The the directory to write the ZenML files to")
@click.argument("path")
def view(path: str, directory: str, out_directory: str):
    """Convert a LookML view to a ZenML view with no other context"""
    result_objects = _convert_logic(path, directory, out_directory, convert_type="view")

    if result_objects:
        click.echo("\n\n".join([json.dumps(o) for o in result_objects]))


@cli_group.command()
@click.option("--directory", default=None, help="The directory of the LookML project to use in conversion")
@click.option("--out-directory", default=None, help="The the directory to write the ZenML files to")
@click.argument("path")
def model(path: str, directory: str, out_directory: str):
    """Convert a LookML model to a ZenML view with no other context"""
    result_objects = _convert_logic(path, directory, out_directory, convert_type="model")
    if result_objects:
        click.echo("\n\n".join([json.dumps(o) for o in result_objects]))


@cli_group.command()
@click.option("--directory", default=None, help="The directory of the LookML project to use in conversion")
@click.option("--out-directory", default=None, help="The the directory to write the ZenML files to")
@click.argument("path")
def dashboard(path: str, directory: str, out_directory: str):
    """Convert a LookML dashboard to a ZenML view with context from the project"""
    result_objects = _convert_logic(path, directory, out_directory, convert_type="dashboard")
    if result_objects:
        click.echo("\n\n".join([json.dumps(o) for o in result_objects]))


@cli_group.command()
@click.option("--out_directory", default=None, help="The the directory to write the ZenML files to")
@click.argument("directory")
def convert(directory: str, out_directory: str):
    """Convert a LookML dashboard to a ZenML view with context from the project"""
    if not out_directory:
        raise ValueError("--out-directory is a required option")
    project = LookMLProjectConverter(in_directory=directory, out_directory=out_directory)
    project.convert()
    click.echo(f"Project converted and saved to {out_directory}")


if __name__ == "__main__":
    cli_group()



import re
import os
import lkml
import json
from ruamel.yaml import YAML
import uuid
from collections import defaultdict
from typing import Dict, Any, List

from lookml_zenml.lookml_models import (
    LookMLProject,
    LookMLModel,
    LookMLView,
    LookMLJoin,
    LookMLExplore,
    LookMLDashboard,
    LookMLDimension,
    LookMLDimensionGroup,
    LookMLMeasure,
    LookMLDashboardElement,
    fields_to_replace,
)

FIELD_KEY_ORDER = ["name", "field_type", "type", "description", "timeframes", "sql"]
UNSUPPORTED_MEASURE_TYPES = [
    "date",
    "date_time",
    "list",
    "median_distinct",
    "percent_of_previous",
    "percent_of_total",
    "percentile_distinct",
    "running_total",
    "period_over_period",
    "yesno",
    "int",
]
UNSUPPORTED_DIMENSION_TYPES = ["zipcode"]
UNSUPPORTED_DIMENSION_GROUP_TYPES = ["date", "datetime"]
yaml = YAML()
yaml.indent(sequence=4, offset=2)
yaml.preserve_quotes = False
yaml.default_style = None


# Configure YAML to use literal block scalars for multiline strings
def represent_str(dumper, data):
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.representer.add_representer(str, represent_str)


class LookMLProjectConverter:
    def __init__(self, in_directory: str = None, out_directory: str = None):
        self.in_directory = in_directory
        self.out_directory = out_directory
        self.field_mappings = []
        self._views = []
        self._models = []
        self._explore_from_mapping = {}
        self._default_model_name = None

    def convert(self):
        if not self.in_directory:
            raise ValueError(
                "You must pass the in_directory argument to the class to use this function. "
                "If you want to convert dict -> dict use the convert_project function in this class"
            )
        lookml_project = self.load(self.in_directory)
        models, views, dashboards, topics = self.convert_project(lookml_project)

        if not self.out_directory:
            raise ValueError(
                "You must pass the out_directory argument to the class to use this function. "
                "If you want to convert dict -> dict use the convert_project function in this class"
            )
        self.dump(self.out_directory, models, views, dashboards, topics)

    def convert_project(self, lookml_project_dict: dict):
        """Convert a raw LookML project dictionary to ZenML format with type validation."""
        models, views, dashboards, topics = [], [], [], []
        view_metadata = {}
        views_to_models = {}

        # Initialize and validate LookML project structure
        try:
            lookml_project = LookMLProject.from_dict(lookml_project_dict)
            # print("Successfully validated complete LookML project structure")
        except Exception as e:
            print(f"Warning: Could not validate full project structure: {e}")
            raise e

        # Process models with validation
        for model_object in lookml_project.models:
            model = self.convert_model(model_object, generate_view_metadata=True)

            model_view_metadata = model.pop("view_metadata", {})
            self._explore_from_mapping[model["name"]] = model_view_metadata["explore_from_mapping"]
            view_metadata.update(model_view_metadata["graph"])
            for view_name in set(model_view_metadata["graph"].keys()):
                views_to_models[view_name] = model["name"]
            models.append(model)

            # Process topics (explores) from each model
            for explore_object in model_object.explores:
                topic = self.convert_topic(explore_object, model["name"])
                topics.append(topic)

        self._models = models

        self._default_model_name = models[0]["name"]
        for view_object in lookml_project.views:
            view = self.convert_view(
                view_object,
                views_to_models.get(view_object.name, self._default_model_name),
                access_filters=[],
                joins=view_metadata.get(view_object.name, []),
            )
            views.append(view)

        self._views = views

        # Process dashboards with validation
        for dashboard_object in lookml_project.dashboards:
            dashboard = self.convert_dashboard(dashboard_object)
            dashboards.append(dashboard)

        return models, views, dashboards, topics

    @staticmethod
    def convert_model(model_object: LookMLModel, generate_view_metadata: bool = False):
        """Convert a LookML model to a ZenML model
        Args: model (dict): A LookML model
              generate_view_metadata (bool): Whether to generate view metadata
                                                based on the explores in the model
        Returns: dict: A ZenML model (if generate_view_metadata is False) or a ZenML
                        with the extra key 'view_metadata' (if generate_view_metadata is True)
        """
        try:
            model_yaml = {
                "version": 1,
                "type": "model",
                "name": model_object.name,
                "connection": model_object.connection,
            }
        except KeyError as e:
            raise KeyError(
                f"The LookML model is missing a required key: {e}. Please check your model and try again."
            )
        if model_object.label:
            model_yaml["label"] = model_object.label
        if model_object.access_grants:
            model_yaml["access_grants"] = [ag.model_dump() for ag in model_object.access_grants]
        if model_object.week_start_day:
            model_yaml["week_start_day"] = model_object.week_start_day.lower()

        if generate_view_metadata:
            model_yaml["view_metadata"] = LookMLProjectConverter.get_view_metadata(model_object)
        return model_yaml

    @staticmethod
    def parse_sql_on(sql_on: str):
        # This code parses a sql_on to determine the relationship between two views
        remaining_sql = sql_on
        references = LookMLProjectConverter.fields_to_replace(sql_on)
        for field in references:
            remaining_sql = remaining_sql.replace(f"${{{field}}}", "")

        # Iff this is true then the sql_on is really just comparing two fields to each other
        if remaining_sql.strip() == "=":
            return "identifier", references

        # Otherwise we need to assume the sql_on is custom
        return "custom", references

    def parse_join(join: LookMLJoin):
        if join.sql_on:
            return LookMLProjectConverter.parse_sql_on(join.sql_on)
        else:
            print(f'Skipping join {join} without "sql_on"')

    @staticmethod
    def get_view_metadata(model: LookMLModel) -> dict:

        explore_from_mapping: Dict[str, Dict[str, str]] = defaultdict(dict)
        identifier_graph: Dict[str, Dict[str, List[Dict[str, Any]]]] = defaultdict(lambda: defaultdict(list))
        for explore in model.explores:
            if explore.from_:
                explore_from_mapping[explore.name][explore.name] = explore.from_.name
            else:
                explore_from_mapping[explore.name][explore.name] = explore.name

            root_view = explore_from_mapping[explore.name][explore.name]
            for join in explore.joins:
                if join.from_:
                    explore_from_mapping[explore.name][join.name] = join.from_
                pair_id = str(uuid.uuid4())
                join_view = join.name if not join.from_ else join.from_

                # Relationship always corresponds to the root view in the explore
                # many_to_one is the default join type
                relationship = join.relationship or "many_to_one"
                join_type, references = LookMLProjectConverter.parse_join(join)

                # If the relationship is one_to_one then we can assume that the
                # key's used in the join are unique (primary) in both views
                if relationship == "one_to_one":
                    if join_type == "identifier":
                        for reference in references:
                            view_name, field_name = reference.split(".")
                            if view_name in explore_from_mapping[explore.name]:
                                view_name = explore_from_mapping[explore.name][view_name]
                            identifier = LookMLProjectConverter._to_identifier(field_name, "primary", pair_id)
                            identifier_graph[view_name][identifier["sql"]].append(identifier)

                # If the relationship is one_to_many then we can assume that the
                # key's used in the join are unique (primary) ONLY if the key references the root view
                elif relationship == "one_to_many":
                    if join_type == "identifier":
                        for reference in references:
                            view_name, field_name = reference.split(".")
                            if view_name == root_view:
                                identifier = LookMLProjectConverter._to_identifier(
                                    field_name, "primary", pair_id
                                )
                            else:
                                identifier = LookMLProjectConverter._to_identifier(
                                    field_name, "foreign", pair_id
                                )
                            if view_name in explore_from_mapping[explore.name]:
                                view_name = explore_from_mapping[explore.name][view_name]
                            identifier_graph[view_name][identifier["sql"]].append(identifier)

                # We do not want to replicate many_to_many joins since they get out of hand quickly
                elif relationship == "many_to_many":
                    pass

                # If the relationship is many_to_one then we can assume that the
                # key's used in the join are unique (primary) ONLY if the key references the join view
                elif relationship == "many_to_one":
                    if join_type == "identifier":
                        for reference in references:
                            view_name, field_name = reference.split(".")
                            if view_name in explore_from_mapping[explore.name]:
                                view_name = explore_from_mapping[explore.name][view_name]
                            if view_name == join_view:
                                identifier = LookMLProjectConverter._to_identifier(
                                    field_name, "primary", pair_id
                                )
                            else:
                                identifier = LookMLProjectConverter._to_identifier(
                                    field_name, "foreign", pair_id
                                )
                            identifier_graph[view_name][identifier["sql"]].append(identifier)
                else:
                    raise NotImplementedError(f"Relationship {relationship} is not supported")

        resolved_id_graph, pairs_resolved = defaultdict(list), []
        for view_name, graph in identifier_graph.items():
            for _, identifiers in graph.items():
                names, view_types, view_sql = [], defaultdict(list), defaultdict(list)

                # Loop through identifiers unique on view -> sql and then find the
                # associated pair to connect for the right naming
                for identifier in identifiers:
                    pair_id = identifier["pair_id"]
                    views, ids = LookMLProjectConverter._get_pair(pair_id, identifier_graph)
                    for v, i in zip(views, ids):
                        if i["pair_id"] not in pairs_resolved:
                            names.append(i["name"])
                            view_types[v].append(i["type"])
                            view_sql[v].append(i["sql"])

                # Mark pairs as resolved
                pairs_resolved.extend(list(set(identifier["pair_id"] for identifier in identifiers)))

                # Names need to match across all identifiers, so we find the longest name in
                # the set of potentially joinable ids and use that name across all of them
                if names:
                    longest_name = max(names, key=len)
                    derived_view_identifier_types = {
                        v: "primary" if "primary" in types else "foreign" for v, types in view_types.items()
                    }
                    derived_view_sql = {v: list(set(sql))[0] for v, sql in view_sql.items()}
                    for view in derived_view_identifier_types.keys():
                        identifier = {
                            "name": longest_name,
                            "type": derived_view_identifier_types[view],
                            "sql": derived_view_sql[view],
                        }
                        if identifier not in resolved_id_graph[view]:
                            resolved_id_graph[view].append(identifier)

        return {"graph": resolved_id_graph, "explore_from_mapping": explore_from_mapping}

    @staticmethod
    def _get_pair(pair_id, id_graph):
        views, ids = [], []
        for view_name, graph in id_graph.items():
            for _, identifiers in graph.items():
                for identifier in identifiers:
                    if identifier["pair_id"] == pair_id:
                        views.append(view_name)
                        ids.append(identifier)
        return views, ids

    @staticmethod
    def convert_view(view: LookMLView, model_name: str, access_filters: list = [], joins: list = []):
        zenml_view = {
            "version": 1,
            "type": "view",
            "name": view.name,
            "model_name": model_name,
            "required_access_grants": view.required_access_grants,
            "sets": [s.model_dump() for s in view.sets],
            "access_filters": [af.model_dump() for af in access_filters],
            "identifiers": joins,
        }

        if view.view_label:
            zenml_view["label"] = view.view_label

        if view.description:
            zenml_view["description"] = view.description

        if view.sql_table_name:
            zenml_view["sql_table_name"] = view.sql_table_name

        if view.derived_table:
            # Convert literal \n characters to actual newlines for better YAML formatting
            sql_content = view.derived_table.sql.replace("\\n", "\n")
            zenml_view["derived_table"] = {"sql": sql_content}

        dimensions = [
            f for field in view.dimensions if (f := LookMLProjectConverter.convert_dimension(field))
        ]
        dimension_groups = [
            f
            for field in view.dimension_groups
            if (f := LookMLProjectConverter.convert_dimension_group(field))
        ]
        measures = [f for field in view.measures if (f := LookMLProjectConverter.convert_measure(field))]

        fields = dimensions + dimension_groups + measures

        first_date_field = next((f["name"] for f in fields if f["field_type"] == "dimension_group"), None)
        if first_date_field:
            zenml_view["default_date"] = first_date_field

        zenml_view["fields"] = fields

        return zenml_view

    def convert_dashboard(self, dashboard_object: LookMLDashboard):
        zenml_data = {
            "version": 1,
            "type": "dashboard",
            "name": dashboard_object.dashboard,
            "label": dashboard_object.title,
            "elements": [],
        }
        if dashboard_object.description:
            zenml_data["description"] = dashboard_object.description

        if dashboard_object.filters:
            zenml_data["filters"] = []
            for f in dashboard_object.filters:
                field = self._get_field(f.field, self._views, model_name=f.model, explore_name=f.explore)
                default_value = f.default_value
                if default_value == "No":
                    default_value = False
                elif default_value == "Yes":
                    default_value = True
                zenml_data["filters"].append({"name": f.name, "field": field["id"], "value": default_value})

        sorted_elements = sorted(dashboard_object.elements, key=lambda x: (x.row, x.col))
        for element in sorted_elements:
            zenml_element = self._translate_dashboard_element(element)
            if zenml_element:
                zenml_data["elements"].append(zenml_element)

        return zenml_data

    def _translate_dashboard_element(self, element: LookMLDashboardElement):
        if element.merged_queries:
            return self._translate_merged_queries_element(element)
        return self._translate_vanilla_element(element)

    def _translate_vanilla_element(self, element: LookMLDashboardElement):
        # We do not support conditional formatting in dashboards
        if not element.model and element.type not in {"button", "text"}:
            return None

        elif element.type == "text":
            model_name = self._default_model_name if self._default_model_name else "todo"
            zenml_element = {"model": model_name, "type": "markdown", "size": "quarter"}
            zenml_element["content"] = element.body_text
            return zenml_element

        elif element.type == "button":
            model_name = self._default_model_name if self._default_model_name else "todo"
            zenml_element = {"model": model_name, "type": "markdown", "size": "quarter"}
            content_json = json.loads(element.rich_content_json)
            link_url = content_json.get("href", "")

            if link_text := content_json.get("text", ""):
                content = f"[{link_text}]({link_url})"
            else:
                content = link_url

            zenml_element["content"] = content
            return zenml_element

        zenml_element = {"model": element.model, "metrics": [], "slice_by": []}
        if element.title:
            zenml_element["title"] = element.title

        # Add fields
        if not element.fields:
            return None

        for f in element.fields:
            field = self._get_field(f, self._views, model_name=element.model, explore_name=element.explore)
            if field["field_type"] == "measure":
                zenml_element["metrics"].append(field["id"])
            else:
                zenml_element["slice_by"].append(field["id"])

        if element.pivots:
            zenml_element["pivot_by"] = []
            for p in element.pivots:
                field = self._get_field(
                    p, self._views, model_name=element.model, explore_name=element.explore
                )
                zenml_element["pivot_by"].append(field["id"])

        if element.filters:
            zenml_element["filters"] = []
            for k, v in element.filters.items():
                field = self._get_field(
                    k, self._views, model_name=element.model, explore_name=element.explore
                )
                if v != "":
                    if v == "No":
                        v = False
                    elif v == "Yes":
                        v = True
                    zenml_element["filters"].append({"field": field["id"].lower(), "value": v})

        if element.listen:
            zenml_element["listen"] = element.listen

        if element.sorts:
            zenml_element["sort"] = []
            for sort_str in element.sorts:
                split_str = sort_str.split(" ")
                field = self._get_field(
                    split_str[0],
                    self._views,
                    model_name=element.model,
                    explore_name=element.explore,
                )
                if len(split_str) == 1:
                    zenml_element["sort"].append({"field": field["id"].lower(), "value": "asc"})
                else:
                    zenml_element["sort"].append(
                        {"field": field["id"].lower(), "value": split_str[1].lower()}
                    )

        if element.width:
            relative_width = float(element.width) / 24
            if relative_width > 0.5:
                zenml_element["size"] = "full"
            elif relative_width > 0.25:
                zenml_element["size"] = "half"
            else:
                zenml_element["size"] = "quarter"

        if element.show_totals:
            zenml_element["show_totals"] = bool(element.show_totals)

        if element.show_value_labels is not None:
            zenml_element["show_annotations"] = bool(element.show_value_labels)

        if element.limit:
            zenml_element["row_limit"] = element.limit

        if element.dynamic_fields:
            zenml_element["table_calculations"] = []
            for dynamic_field in element.dynamic_fields:
                # We do not currently support dynamic fields that are not table calculations
                if dynamic_field.get("category") == "table_calculation" and not dynamic_field.get(
                    "is_disabled", False
                ):
                    zenml_element["table_calculations"].append(
                        {
                            "title": dynamic_field["label"],
                            "formula": self._clean_table_calc(dynamic_field["expression"], element=element),
                            "format": dynamic_field.get("value_format_name", "decimal_1"),
                        }
                    )

        plot_lookup = {
            "looker_area": "area",
            "looker_donut_multiples": "donut",
            "looker_bar": "horizontal_bar",
            "looker_pie": "pie",
            "table": "table_only",
            "looker_scatter": "scatter",
            "single_value": "stat_plot",
        }

        if element.type == "looker_column":
            plot = "grouped_bar" if len(zenml_element["slice_by"]) > 1 else "bar"
            if element.stacking == "normal":
                plot_options = {"grouped_bar": {"display_type": "STACKED"}}
                zenml_element["plot_options"] = plot_options
            zenml_element["plot_type"] = plot
        elif element.type == "looker_line":
            "line" or "multi_line"
            plot = "multi_line" if len(zenml_element["slice_by"]) > 1 else "line"
            zenml_element["plot_type"] = plot
        elif element.type in plot_lookup:
            zenml_element["plot_type"] = plot_lookup[element.type]
            if zenml_element["plot_type"] == "table_only":
                zenml_element["force_table"] = True

        return zenml_element

    def _get_field(self, field_id: str, views: list, model_name: str = None, explore_name: str = None):
        if len(field_id.split(".")) == 2 and model_name and explore_name:
            explore_view_name, field_name = field_id.split(".")
            if explore_view_name in self._explore_from_mapping.get(model_name, {}).get(explore_name, {}):
                explore_view_name = self._explore_from_mapping[model_name][explore_name][explore_view_name]
            field_id = f"{explore_view_name}.{field_name}"

        for v in views:
            for f in v["fields"]:
                if f"{v['name']}.{f['name']}" == field_id.lower() and f["field_type"] != "dimension_group":
                    return {**f, "id": field_id}
                elif f"{v['name']}.{f['name']}" in field_id.lower() and f["field_type"] == "dimension_group":
                    return {**f, "id": field_id}
        # If we can't find the field assume it will be a dimension on a dashboard
        return {"field_type": "dimension", "id": field_id}

    def _clean_table_calc(self, formula: str, element: LookMLDashboardElement):
        for field_to_replace in self.fields_to_replace(formula):
            if ":row_total" in field_to_replace:
                field_id = field_to_replace.split(":row_total")[0]
                field = self._get_field(
                    field_id,
                    self._views,
                    model_name=element.model,
                    explore_name=element.explore,
                )
                formula = formula.replace(f"${{{field_to_replace}}}", f"sum([{field['id']}])")
            else:
                field = self._get_field(
                    field_to_replace,
                    self._views,
                    model_name=element.model,
                    explore_name=element.explore,
                )
                formula = formula.replace(f"${{{field_to_replace}}}", f"[{field['id']}]")
        return formula

    def _translate_merged_queries_element(self, element: LookMLDashboardElement):
        # We do not support translation for looker filter expressions

        filters_by_expression = defaultdict(list)
        for merged in element.merged_queries:
            if "join_fields" in merged:
                for join_field in merged["join_fields"]:
                    self.field_mappings.append([join_field["field_name"], join_field["source_field_name"]])
            if "filters" in merged:
                for f, e in merged["filters"].items():
                    filters_by_expression[e].append(f)

        for filter_fields in filters_by_expression.values():
            if len(filter_fields) > 1:
                self.field_mappings.append([filter_fields[0], filter_fields[1]])

        running_fields = []
        for merged in element.merged_queries:
            for f in merged["fields"]:
                # Add fields that are not mappings
                matched_pair = next((pair for pair in self.field_mappings if f in pair), None)

                # We can add one and only one field from the mappings to the running fields
                if matched_pair and all(p not in running_fields for p in matched_pair):
                    running_fields.append(f)

                # We can always add fields that are not in the mappings at all
                if all(f not in pair for pair in self.field_mappings):
                    running_fields.append(f)

        first_element = element.merged_queries[0]
        first_element["name"] = element.name
        first_element["fields"] = running_fields

        return self._translate_vanilla_element(LookMLDashboardElement.from_dict(first_element))

    @staticmethod
    def fields_to_replace(text: str):
        return fields_to_replace(text)

    def load(self, in_directory: str):
        lookml = {"views": [], "models": [], "dashboards": []}
        for root, _, files in os.walk(in_directory):
            for file in files:
                if self.is_lookml_file(file):
                    lookml_file = os.path.join(root, file)
                    with open(lookml_file) as f:
                        lookml_dict = lkml.load(f)

                    if ".model." in file:
                        lookml_dict["name"] = file.split(".model.")[0]
                        lookml["models"].append(lookml_dict)
                    elif ".view." in file:
                        lookml["views"].extend(lookml_dict["views"])
                    else:
                        raise NotImplementedError(f"File {file} is not a model or view")

                if self.is_lookml_dashboard_file(file):
                    lookml_file = os.path.join(root, file)
                    with open(lookml_file) as f:
                        lookml_dict = yaml.load(f)
                    lookml["dashboards"].extend(lookml_dict)
        return lookml

    @staticmethod
    def dump(out_directory: str, models: list, views: list, dashboards: list, topics: list):
        zenlytic_project = {
            "name": "new_zenlytic_project",
            "profile": "new_zenlytic_project",
            "dashboard-paths": ["dashboards"],
            "view-paths": ["views"],
            "model-paths": ["models"],
            "topic-paths": ["topics"],
        }
        if not os.path.exists(out_directory):
            raise ValueError(f"The output directory you specified, `{out_directory}` does not exist.")

        with open(os.path.join(out_directory, "zenlytic_project.yml"), "w") as f:
            yaml.dump(zenlytic_project, f)

        for directory in ["models", "views", "dashboards", "topics"]:
            fully_qualified_path = os.path.join(out_directory, directory)
            if not os.path.exists(fully_qualified_path):
                os.mkdir(fully_qualified_path)

        for model in models:
            with open(os.path.join(out_directory, f"models/{model['name']}_model.yml"), "w") as f:
                yaml.dump(model, f)

        for view in views:
            with open(os.path.join(out_directory, f"views/{view['name']}_view.yml"), "w") as f:
                yaml.dump(view, f)

        for dashboard in dashboards:
            with open(os.path.join(out_directory, f"dashboards/{dashboard['name']}.yml"), "w") as f:
                yaml.dump(dashboard, f)

        for topic in topics:
            with open(os.path.join(out_directory, f"topics/{topic['name']}_topic.yml"), "w") as f:
                topic.pop("name")
                yaml.dump(topic, f)

    @staticmethod
    def convert_dimension(dimension: LookMLDimension):
        converted = {
            "name": dimension.name,
            "field_type": "dimension",
        }
        if dimension.sql:
            converted["sql"] = dimension.sql.replace("\\n", "\n")
        elif dimension.case:
            converted["case"] = dimension.case
        else:
            converted["sql"] = "${TABLE}." + dimension.name
        if dimension.label:
            converted["label"] = dimension.label
        if dimension.required_access_grants:
            converted["required_access_grants"] = dimension.required_access_grants
        if dimension.group_label:
            converted["group_label"] = dimension.group_label
        if dimension.description:
            converted["description"] = dimension.description
        if dimension.hidden:
            converted["hidden"] = dimension.hidden == "yes"
        if dimension.primary_key:
            converted["primary_key"] = dimension.primary_key == "yes"
        if dimension.tiers:
            converted["tiers"] = [int(tier) for tier in dimension.tiers]
        if dimension.type and dimension.type not in UNSUPPORTED_DIMENSION_TYPES:
            converted["type"] = dimension.type
        else:
            converted["type"] = "string"

        if converted["type"] == "string":
            converted["searchable"] = True

        if dimension.links:
            converted["link"] = dimension.links[0].url

        return LookMLProjectConverter.sort_dict(converted, FIELD_KEY_ORDER)

    @staticmethod
    def convert_dimension_group(dimension_group: LookMLDimensionGroup):
        converted = {"name": dimension_group.name, "field_type": "dimension_group"}
        if dimension_group.label:
            converted["label"] = dimension_group.label
        if dimension_group.description:
            converted["description"] = dimension_group.description
        if dimension_group.group_label:
            converted["group_label"] = dimension_group.group_label
        if dimension_group.required_access_grants:
            converted["required_access_grants"] = dimension_group.required_access_grants
        if dimension_group.type and dimension_group.type not in UNSUPPORTED_DIMENSION_GROUP_TYPES:
            converted["type"] = dimension_group.type
        else:
            converted["type"] = "time"
        if dimension_group.sql:
            converted["sql"] = dimension_group.sql.replace("\\n", "\n")
        if dimension_group.sql_start:
            converted["sql_start"] = dimension_group.sql_start.replace("\\n", "\n")
        if dimension_group.sql_end:
            converted["sql_end"] = dimension_group.sql_end.replace("\\n", "\n")
        if dimension_group.hidden:
            converted["hidden"] = dimension_group.hidden == "yes"
        if dimension_group.primary_key:
            converted["primary_key"] = converted["primary_key"] == "yes"
        if dimension_group.convert_tz:
            converted["convert_tz"] = dimension_group.convert_tz == "yes"
        if dimension_group.convert_timezone:
            converted["convert_timezone"] = dimension_group.convert_timezone == "yes"
        if dimension_group.type == "time" and not dimension_group.timeframes:
            converted["timeframes"] = ["raw", "date", "week", "month", "quarter", "year"]
        if dimension_group.timeframes:
            converted["timeframes"] = dimension_group.timeframes
        if dimension_group.intervals:
            converted["intervals"] = dimension_group.intervals
        elif dimension_group.type == "duration" and not dimension_group.intervals:
            converted["intervals"] = ["day", "week", "month", "quarter", "year"]

        return LookMLProjectConverter.sort_dict(converted, FIELD_KEY_ORDER)

    @staticmethod
    def convert_measure(measure: LookMLMeasure):
        # These measure types are not supported in LookML -> ZenML conversion
        if measure.type and measure.type in UNSUPPORTED_MEASURE_TYPES:
            return {}

        converted = {
            "name": measure.name,
            "field_type": "measure",
            "sql": measure.sql.replace("\\n", "\n") if measure.sql else "",
        }
        if measure.label:
            converted["label"] = measure.label
        if measure.type:
            converted["type"] = measure.type
        if measure.required_access_grants:
            converted["required_access_grants"] = measure.required_access_grants
        if measure.sql_distinct_key:
            converted["sql_distinct_key"] = measure.sql_distinct_key.replace("\\n", "\n")
        if measure.description:
            converted["description"] = measure.description
        if measure.hidden:
            converted["hidden"] = measure.hidden == "yes"
        if measure.percentile:
            converted["percentile"] = measure.percentile
        if measure.group_label:
            converted["group_label"] = measure.group_label

        if measure.value_format_name:
            converted["value_format_name"] = measure.value_format_name

        if measure.filters:
            converted["filters"] = [f.model_dump() for f in measure.filters]

        return LookMLProjectConverter.sort_dict(converted, FIELD_KEY_ORDER)

    @staticmethod
    def convert_topic(explore_object: LookMLExplore, model_name: str):
        """Convert a LookML explore to a ZenML topic."""
        topic = {
            "version": 1,
            "type": "topic",
            "name": explore_object.name,
            "model_name": model_name,
        }

        if explore_object.label:
            topic["label"] = explore_object.label
        else:
            topic["label"] = explore_object.name.replace("_", " ").title()
        if explore_object.description:
            topic["description"] = explore_object.description
        if explore_object.hidden is not None:
            topic["hidden"] = explore_object.hidden == "yes"
        if explore_object.from_:
            topic["base_view"] = explore_object.from_.name
        elif explore_object.view_name:
            topic["base_view"] = explore_object.view_name
        else:
            topic["base_view"] = explore_object.name
        if explore_object.extends:
            raise NotImplementedError("Extends are not supported in ZenML")

        if explore_object.extension:
            raise NotImplementedError("Extension is not supported in ZenML")

        if explore_object.fields:
            print(
                f"Field restriction is not supported in ZenML in a topic, passing on "
                f"this argument in the explore {explore_object.name}"
            )
        if explore_object.sql_always_where:
            print(
                f"SQL always where is not supported in ZenML, passing on "
                f"this argument in the explore {explore_object.name}"
            )
        if explore_object.required_access_grants:
            topic["required_access_grants"] = explore_object.required_access_grants

        if explore_object.always_filter:
            topic["always_filter"] = explore_object.always_filter.model_dump()["filters"]

        if explore_object.conditionally_filter:
            print(
                f"Conditionally filter is not supported in ZenML, passing on "
                f"this argument in the explore {explore_object.name}"
            )
        if explore_object.access_filter:
            topic["access_filter"] = [af.model_dump() for af in explore_object.access_filter]

        if explore_object.joins:
            topic["views"] = {
                join.name: {
                    "join": {
                        "join_type": join.type or "left_outer",
                        "relationship": join.relationship or "many_to_one",
                        "sql_on": join.sql_on,
                    }
                }
                for join in explore_object.joins
            }

        if explore_object.sql_table_name:
            print(
                f"SQL table name is not supported in ZenML in an explore, passing on "
                f"this argument in the explore {explore_object.name}"
            )

        return topic

    @staticmethod
    def _to_identifier(field_name: str, type: str, pair_id: str):
        return {"name": field_name, "type": type, "sql": f"${{{field_name}}}", "pair_id": pair_id}

    @staticmethod
    def is_lookml_file(file):
        suffixes = [".model.lkml", ".view.lkml", ".model.lookml", ".view.lookml"]
        return any(file.endswith(suffix) for suffix in suffixes)

    @staticmethod
    def is_lookml_dashboard_file(file):
        suffixes = [".dashboard.lkml", ".dashboard.lookml"]
        return any(file.endswith(suffix) for suffix in suffixes)

    @staticmethod
    def sort_dict(d: dict, sort_order: list):
        sorted_dict = {k: d[k] for k in sort_order if k in d}
        sorted_dict.update({k: v for k, v in d.items() if k not in sorted_dict})
        return sorted_dict



import re
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Union, Any, Literal


def fields_to_replace(text: str) -> List[str]:
    if text is None:
        return []
    matches = re.finditer(r"\$\{(.*?)\}", text, re.MULTILINE)
    return [match.group(1) for match in matches]


# Pydantic models for LookML structures
class LookMLDerivedTable(BaseModel):
    """Pydantic model for LookML derived table."""

    sql: str
    sql_trigger_value: Optional[str] = None
    datagroup_trigger: Optional[str] = None
    persist_for: Optional[str] = None
    indexes: Optional[List[str]] = None
    distribution_style: Optional[str] = None
    sortkeys: Optional[List[str]] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDerivedTable":
        """Create LookMLDerivedTable from dictionary."""
        data = data.copy()
        if "explore_source" in data:
            raise NotImplementedError(
                f"explore_source is not implemented as a valid view data source. "
                f"Please remove the file and try conversion again. data: {data}"
            )
        return cls(**data)


class LookMLLink(BaseModel):
    """Pydantic model for LookML link."""

    url: str
    name: Optional[str] = None


class LookMLDimension(BaseModel):
    """Pydantic model for LookML dimension."""

    name: str
    type: Optional[str] = "string"
    sql: Optional[str] = None
    description: Optional[str] = None
    label: Optional[str] = None
    view_label: Optional[str] = None
    group_label: Optional[str] = None
    value_format: Optional[str] = None
    value_format_name: Optional[str] = None
    primary_key: Optional[str] = None
    hidden: Optional[str] = None
    drill_fields: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    required_access_grants: Optional[List[str]] = None
    links: Optional[List[LookMLLink]] = None
    tiers: Optional[List[int]] = None
    case: Optional[Dict[str, Any]] = None
    alpha_sort: Optional[bool] = None
    order_by_field: Optional[str] = None
    suggest_dimension: Optional[str] = None
    suggest_explore: Optional[str] = None
    suggestions: Optional[List[str]] = None
    bypass_suggest_restrictions: Optional[bool] = None
    full_suggestions: Optional[bool] = None
    can_filter: Optional[bool] = None
    fanout_on: Optional[str] = None
    sql_latitude: Optional[str] = None
    sql_longitude: Optional[str] = None
    map_layer_name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDimension":
        """Create LookMLDimension from dictionary."""
        # Parse links if present
        if "links" in data and data["links"]:
            parsed_links = []
            for link in data["links"]:
                if isinstance(link, dict):
                    parsed_links.append(LookMLLink.from_dict(link))
                else:
                    parsed_links.append(LookMLLink(url=link))
            data["links"] = parsed_links

        return cls(**data)


class LookMLDimensionGroup(BaseModel):
    """Pydantic model for LookML dimension group."""

    name: str
    type: Literal["time", "date", "datetime", "duration"] = "time"
    primary_key: Optional[str] = None
    sql: Optional[str] = None
    description: Optional[str] = None
    label: Optional[str] = None
    view_label: Optional[str] = None
    group_label: Optional[str] = None
    timeframes: Optional[List[str]] = None
    intervals: Optional[List[str]] = None
    convert_tz: Optional[str] = None
    convert_timezone: Optional[str] = None
    datatype: Optional[str] = None
    sql_start: Optional[str] = None
    sql_end: Optional[str] = None
    hidden: Optional[bool] = None
    drill_fields: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    required_access_grants: Optional[List[str]] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDimensionGroup":
        """Create LookMLDimensionGroup from dictionary."""
        return cls(**data)


class LookMLFilter(BaseModel):
    """Pydantic model for LookML filter."""

    field: str
    value: str

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLFilter":
        """Create LookMLFilter from dictionary."""
        return cls(**data)

    @classmethod
    def process_list_of_filters(
        cls, filters: List[Dict[str, Any]], alias_to_view_mapping: Dict[str, str] = {}
    ) -> List["LookMLFilter"]:
        """Process a list of filters and return a list of LookMLFilter objects."""

        filter_objects = []
        for f in filters:
            if isinstance(f, list):
                for sub_filter in f:
                    for field, value in sub_filter.items():
                        field = LookMLFilter.replace_field_alias_with_view(field, alias_to_view_mapping)
                        filter_objects.append(cls(field=field.lower(), value=value))
            else:
                field = LookMLFilter.replace_field_alias_with_view(f["field"], alias_to_view_mapping)
                filter_objects.append(cls(field=field.lower(), value=f["value"]))

        return filter_objects

    @staticmethod
    def replace_field_alias_with_view(field_id: str, alias_to_view_mapping: Dict[str, str]) -> str:
        """Replace a field alias with the corresponding view name."""
        if "." in field_id:
            view_name = field_id.split(".")[0]
            field_name = field_id.split(".")[1]
            if view_name in alias_to_view_mapping:
                view_name = alias_to_view_mapping[view_name]
                field_id = view_name + "." + field_name
        return field_id


class LookMLMeasure(BaseModel):
    """Pydantic model for LookML measure."""

    name: str
    type: str = "count"
    sql: Optional[str] = None
    description: Optional[str] = None
    label: Optional[str] = None
    view_label: Optional[str] = None
    group_label: Optional[str] = None
    value_format: Optional[str] = None
    value_format_name: Optional[str] = None
    hidden: Optional[str] = None
    drill_fields: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    required_access_grants: Optional[List[str]] = None
    filters: Optional[List[LookMLFilter]] = None
    sql_distinct_key: Optional[str] = None
    percentile: Optional[int] = None
    approximate: Optional[bool] = None
    approximate_threshold: Optional[int] = None
    direction: Optional[str] = None
    can_filter: Optional[bool] = None
    link: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLMeasure":
        """Create LookMLMeasure from dictionary with nested filter conversion."""
        # Convert nested filters
        if "filters__all" in data and data["filters__all"]:
            data = data.copy()
            filters = data.pop("filters__all")
            data["filters"] = LookMLFilter.process_list_of_filters(filters)

        return cls(**data)


class LookMLSet(BaseModel):
    """Pydantic model for LookML set."""

    name: str
    fields: List[str]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLSet":
        """Create LookMLSet from dictionary."""
        return cls(**data)


class LookMLAccessFilter(BaseModel):
    """Pydantic model for LookML access filter."""

    field: str
    user_attribute: str

    @classmethod
    def from_dict(
        cls, data: Dict[str, Any], alias_to_view_mapping: Dict[str, str] = {}
    ) -> "LookMLAccessFilter":
        """Create LookMLAccessFilter from dictionary."""
        data = data.copy()
        if "field" in data:
            data["field"] = LookMLFilter.replace_field_alias_with_view(data["field"], alias_to_view_mapping)
        return cls(**data)


class LookMLView(BaseModel):
    """Pydantic model for LookML view."""

    name: str
    view_label: Optional[str] = None
    description: Optional[str] = None
    sql_table_name: Optional[str] = None
    derived_table: Optional[LookMLDerivedTable] = None
    required_access_grants: Optional[List[str]] = []
    sets: Optional[List[LookMLSet]] = []
    dimensions: Optional[List[LookMLDimension]] = []
    dimension_groups: Optional[List[LookMLDimensionGroup]] = []
    measures: Optional[List[LookMLMeasure]] = []
    filters: Optional[List[Dict[str, Any]]] = []
    parameters: Optional[List[Dict[str, Any]]] = []

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLView":
        """Create LookMLView from dictionary with nested type conversion."""
        data = data.copy()

        # Convert nested derived_table
        if "derived_table" in data and data["derived_table"]:
            try:
                data["derived_table"] = LookMLDerivedTable.from_dict(data["derived_table"])
            except NotImplementedError as e:
                raise NotImplementedError(f"In the view {data['name']}, {str(e)}")

        # Convert nested sets
        if "sets" in data and data["sets"]:
            data["sets"] = [LookMLSet.from_dict(s) if isinstance(s, dict) else s for s in data["sets"]]

        # Convert nested dimensions
        if "dimensions" in data and data["dimensions"]:
            data["dimensions"] = [
                LookMLDimension.from_dict(d) if isinstance(d, dict) else d for d in data["dimensions"]
            ]

        # Convert nested dimension_groups
        if "dimension_groups" in data and data["dimension_groups"]:
            data["dimension_groups"] = [
                LookMLDimensionGroup.from_dict(dg) if isinstance(dg, dict) else dg
                for dg in data["dimension_groups"]
            ]

        # Convert nested measures
        if "measures" in data and data["measures"]:
            data["measures"] = [
                LookMLMeasure.from_dict(m) if isinstance(m, dict) else m for m in data["measures"]
            ]

        return cls(**data)


class LookMLJoin(BaseModel):
    """Pydantic model for LookML join."""

    name: str
    type: Optional[str] = "left_outer"
    relationship: Optional[str] = "many_to_one"
    sql_on: Optional[str] = None
    from_: Optional[str] = Field(None, alias="from")
    view_label: Optional[str] = None
    fields: Optional[List[str]] = None
    required_joins: Optional[List[str]] = None
    foreign_key: Optional[str] = None
    sql_table_name: Optional[str] = None
    sql_where: Optional[str] = None
    required_access_grants: Optional[List[str]] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any], alias_to_view_mapping: Dict[str, str] = {}) -> "LookMLJoin":
        """Create LookMLJoin from dictionary."""
        data = data.copy()
        if "from" in data and data["from"] != data["name"]:
            alias_to_view_mapping[data["name"]] = data["from"]
        if "sql_on" in data:
            for f in fields_to_replace(data["sql_on"]):
                new_field = LookMLFilter.replace_field_alias_with_view(f, alias_to_view_mapping)
                data["sql_on"] = data["sql_on"].replace("${" + f + "}", "${" + new_field + "}")
        if "sql_where" in data:
            for f in fields_to_replace(data["sql_where"]):
                new_field = LookMLFilter.replace_field_alias_with_view(f, alias_to_view_mapping)
                data["sql_where"] = data["sql_where"].replace("${" + f + "}", "${" + new_field + "}")
        if "sql_where" in data and "sql_on" in data:
            data["sql_on"] = data["sql_on"] + " AND " + data["sql_where"]
        return cls(**data)


class LookMLAlwaysFilter(BaseModel):
    """Pydantic model for LookML always filter."""

    filters: List[LookMLFilter]

    @classmethod
    def from_dict(
        cls, data: Dict[str, Any], alias_to_view_mapping: Dict[str, str] = {}
    ) -> "LookMLAlwaysFilter":
        """Create LookMLAlwaysFilter from dictionary with nested filter conversion."""
        data = data.copy()

        # Convert nested filters
        if "filters__all" in data and data["filters__all"]:
            filters = data.pop("filters__all")
            data["filters"] = LookMLFilter.process_list_of_filters(filters, alias_to_view_mapping)

        return cls(**data)


class LookMLConditionallyFilter(BaseModel):
    """Pydantic model for LookML conditionally filter."""

    filters: List[LookMLFilter]
    unless: Optional[List[str]] = None

    @classmethod
    def from_dict(
        cls, data: Dict[str, Any], alias_to_view_mapping: Dict[str, str] = {}
    ) -> "LookMLConditionallyFilter":
        """Create LookMLConditionallyFilter from dictionary with nested filter conversion."""
        data = data.copy()

        if "filters__all" in data and data["filters__all"]:
            filters = data.pop("filters__all")
            data["filters"] = LookMLFilter.process_list_of_filters(filters, alias_to_view_mapping)

        return cls(**data)


class LookMLExploreFrom(BaseModel):
    """Pydantic model for the LookML explore 'from' field.

    This represents the base view that an explore is built from.
    """

    name: str
    view_label: Optional[str] = None
    label: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Union[Dict[str, Any], str]) -> "LookMLExploreFrom":
        """Create LookMLExploreFrom from dictionary."""

        if isinstance(data, dict):
            data = data.copy()
            return cls(**data)
        else:
            return cls(name=data)


class LookMLExplore(BaseModel):
    """Pydantic model for LookML explore."""

    name: str
    description: Optional[str] = None
    label: Optional[str] = None
    group_label: Optional[str] = None
    view_label: Optional[str] = None
    from_: Optional[LookMLExploreFrom] = Field(None, alias="from")
    view_name: Optional[str] = None
    extends: Optional[List[str]] = None
    extension: Optional[str] = None
    hidden: Optional[str] = None
    fields: Optional[List[str]] = None
    sql_always_where: Optional[str] = None
    required_access_grants: Optional[List[str]] = None
    always_filter: Optional[LookMLAlwaysFilter] = None
    conditionally_filter: Optional[LookMLConditionallyFilter] = None
    access_filter: Optional[List[LookMLAccessFilter]] = None
    always_join: Optional[List[str]] = None
    joins: Optional[List[LookMLJoin]] = []
    persist_for: Optional[str] = None
    persist_with: Optional[str] = None
    case_sensitive: Optional[bool] = None
    sql_table_name: Optional[str] = None
    cancel_grouping_fields: Optional[List[str]] = None
    symmetric_aggregates: Optional[bool] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLExplore":
        """Create LookMLExplore from dictionary with nested type conversion."""
        data = data.copy()

        alias_to_view_mapping = {}
        if "view_name" in data and "name" in data:
            alias_to_view_mapping[data["name"]] = data["view_name"]

        # Convert nested from field
        if "from" in data and "name" in data:
            data["from"] = LookMLExploreFrom.from_dict(data["from"])
            alias_to_view_mapping[data["name"]] = data["from"].name

        # Convert nested joins
        if "joins" in data and data["joins"]:
            data["joins"] = [
                LookMLJoin.from_dict(j, alias_to_view_mapping) if isinstance(j, dict) else j
                for j in data["joins"]
            ]

        # Convert nested always_filter
        if "always_filter" in data and data["always_filter"]:
            data["always_filter"] = LookMLAlwaysFilter.from_dict(data["always_filter"], alias_to_view_mapping)

        # Convert nested conditionally_filter
        if "conditionally_filter" in data and data["conditionally_filter"]:
            data["conditionally_filter"] = LookMLConditionallyFilter.from_dict(
                data["conditionally_filter"], alias_to_view_mapping
            )

        # Convert nested access_filter
        if "access_filter" in data and data["access_filter"]:
            data["access_filter"] = [
                LookMLAccessFilter.from_dict(af, alias_to_view_mapping) if isinstance(af, dict) else af
                for af in data["access_filter"]
            ]

        return cls(**data)


class LookMLDatagroup(BaseModel):
    """Pydantic model for LookML datagroup."""

    name: str
    sql_trigger: str
    max_cache_age: str

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDatagroup":
        """Create LookMLDatagroup from dictionary."""
        return cls(**data)


class LookMLAccessGrant(BaseModel):
    """Pydantic model for LookML access grant."""

    name: str
    user_attribute: str
    allowed_values: List[str]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLAccessGrant":
        """Create LookMLAccessGrant from dictionary."""
        return cls(**data)


class LookMLNamedValueFormat(BaseModel):
    """Pydantic model for LookML named value format."""

    name: str
    value_format: str

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLNamedValueFormat":
        """Create LookMLNamedValueFormat from dictionary."""
        return cls(**data)


class LookMLMapLayer(BaseModel):
    """Pydantic model for LookML map layer."""

    name: str
    file: Optional[str] = None
    url: Optional[str] = None
    format: Optional[str] = None
    projection: Optional[str] = None
    extents_json_url: Optional[str] = None
    feature_key: Optional[str] = None
    property_key: Optional[str] = None
    property_label_key: Optional[str] = None
    label: Optional[str] = None
    max_zoom_level: Optional[int] = None
    min_zoom_level: Optional[int] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLMapLayer":
        """Create LookMLMapLayer from dictionary."""
        return cls(**data)


class LookMLTest(BaseModel):
    """Pydantic model for LookML test."""

    name: str
    explore_source: Dict[str, Any]
    assert_: List[Dict[str, Any]] = Field(alias="assert")

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLTest":
        """Create LookMLTest from dictionary."""
        return cls(**data)


class LookMLModel(BaseModel):
    """Pydantic model for LookML model."""

    name: str
    connection: str
    label: Optional[str] = None
    include: Optional[List[str]] = None
    fiscal_month_offset: Optional[int] = None
    persist_for: Optional[str] = None
    persist_with: Optional[str] = None
    case_sensitive: Optional[bool] = None
    week_start_day: Optional[str] = None
    datagroups: Optional[List[LookMLDatagroup]] = None
    access_grants: Optional[List[LookMLAccessGrant]] = []
    named_value_formats: Optional[List[LookMLNamedValueFormat]] = []
    map_layers: Optional[List[LookMLMapLayer]] = []
    tests: Optional[List[LookMLTest]] = []
    explores: Optional[List[LookMLExplore]] = []

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLModel":
        """Create LookMLModel from dictionary with nested type conversion."""
        data = data.copy()

        # Convert nested datagroups
        if "datagroups" in data and data["datagroups"]:
            data["datagroups"] = [
                LookMLDatagroup.from_dict(dg) if isinstance(dg, dict) else dg for dg in data["datagroups"]
            ]

        # Convert nested access_grants
        if "access_grants" in data and data["access_grants"]:
            data["access_grants"] = [
                LookMLAccessGrant.from_dict(ag) if isinstance(ag, dict) else ag
                for ag in data["access_grants"]
            ]

        # Convert nested named_value_formats
        if "named_value_formats" in data and data["named_value_formats"]:
            data["named_value_formats"] = [
                LookMLNamedValueFormat.from_dict(nvf) if isinstance(nvf, dict) else nvf
                for nvf in data["named_value_formats"]
            ]

        # Convert nested map_layers
        if "map_layers" in data and data["map_layers"]:
            data["map_layers"] = [
                LookMLMapLayer.from_dict(ml) if isinstance(ml, dict) else ml for ml in data["map_layers"]
            ]

        # Convert nested tests
        if "tests" in data and data["tests"]:
            data["tests"] = [LookMLTest.from_dict(t) if isinstance(t, dict) else t for t in data["tests"]]

        # Convert nested explores
        if "explores" in data and data["explores"]:
            data["explores"] = [
                LookMLExplore.from_dict(e) if isinstance(e, dict) else e for e in data["explores"]
            ]

        return cls(**data)


class LookMLDashboardFilter(BaseModel):
    """Pydantic model for LookML dashboard filter."""

    name: str
    field: str
    default_value: Optional[Union[str, bool, int, float]] = None
    model: Optional[str] = None
    explore: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDashboardFilter":
        """Create LookMLDashboardFilter from dictionary."""
        return cls(**data)


class LookMLDashboardElement(BaseModel):
    """Pydantic model for LookML dashboard element."""

    name: str
    title: Optional[str] = None
    model: Optional[str] = None
    explore: Optional[str] = None
    type: Optional[str] = None
    fields: Optional[List[str]] = None
    pivots: Optional[List[str]] = None
    filters: Optional[Dict[str, str]] = None
    sorts: Optional[List[str]] = None
    limit: Optional[int] = None
    width: Optional[int] = None
    height: Optional[int] = None
    row: Optional[int] = 0
    col: Optional[int] = 0
    stacking: Optional[str] = None
    listen: Optional[Dict[str, str]] = None
    show_totals: Optional[bool] = None
    show_value_labels: Optional[bool] = None
    dynamic_fields: Optional[List[Dict[str, Any]]] = None
    merged_queries: Optional[List[Dict[str, Any]]] = None
    body_text: Optional[str] = None
    rich_content_json: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDashboardElement":
        """Create LookMLDashboardElement from dictionary."""
        return cls(**data)


class LookMLDashboard(BaseModel):
    """Pydantic model for LookML dashboard."""

    dashboard: str
    title: str
    description: Optional[str] = None
    layout: Optional[str] = None
    preferred_viewer: Optional[str] = None
    filters: Optional[List[LookMLDashboardFilter]] = None
    elements: Optional[List[LookMLDashboardElement]] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLDashboard":
        """Create LookMLDashboard from dictionary with nested type conversion."""
        data = data.copy()

        # Convert nested filters
        if "filters" in data and data["filters"]:
            data["filters"] = [
                LookMLDashboardFilter.from_dict(f) if isinstance(f, dict) else f for f in data["filters"]
            ]

        # Convert nested elements
        if "elements" in data and data["elements"]:
            data["elements"] = [
                LookMLDashboardElement.from_dict(e) if isinstance(e, dict) else e for e in data["elements"]
            ]

        return cls(**data)


class LookMLProject(BaseModel):
    """Pydantic model for complete LookML project."""

    models: List[LookMLModel]
    views: List[LookMLView]
    dashboards: List[LookMLDashboard]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookMLProject":
        """Create LookMLProject from dictionary with nested type conversion."""
        data = data.copy()

        # Convert nested models
        if "models" in data and data["models"]:
            data["models"] = [LookMLModel.from_dict(m) if isinstance(m, dict) else m for m in data["models"]]

        # Convert nested views
        if "views" in data and data["views"]:
            data["views"] = [LookMLView.from_dict(v) if isinstance(v, dict) else v for v in data["views"]]

        # Convert nested dashboards
        if "dashboards" in data and data["dashboards"]:
            data["dashboards"] = [
                LookMLDashboard.from_dict(d) if isinstance(d, dict) else d for d in data["dashboards"]
            ]

        return cls(**data)
